import ij.ImagePlus;
import ij.ImageStack;
import ij.gui.GenericDialog;
import ij.gui.Plot;
import ij.plugin.filter.PlugInFilter;
import ij.process.Blitter;
import ij.process.FloatProcessor;
import ij.process.ImageProcessor;
import snakeprogram.MultipleSnakesStore;
import snakeprogram.Snake;
import snakeprogram.SnakeIO;
import speckles.SpeckleWriter;

import java.awt.*;
import java.io.BufferedWriter;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Random;
public class LE_Radius implements PlugInFilter {
			
    ImagePlus imp;
    double LINE_WIDTH;
    double front;
    double back;
    double centerX;
    double centerY;
    double radius;
    double dx,dt;
 	double[][] rads;
 
 	int nstks,dtheta, tfit;
    ArrayList<ImageProcessor> out_stack = new ArrayList<ImageProcessor>();
    ImageProcessor new_processor;
    ArrayList<double[]>kymo_pixels = new ArrayList<double[]>(360);
    BufferedWriter bw;
    
    public int setup(String arg, ImagePlus imp) {
		this.imp = imp;
		return DOES_ALL;
	}

    public void run(ImageProcessor ip) {
    	int maxF=0;
    	// read in the snake file
        MultipleSnakesStore store = SnakeIO.loadSnakes((Frame)null, new HashMap<String, Double>());
        Snake s = store.getLastSnake();
        ImageProcessor improc;
        ArrayList<double[]> cnets;
        Plot showing = null;
        ImageStack istack = imp.getStack();
        
        int fs=0;
        centerX = 0;
        centerY = 0;
        radius = 0;
        
        for(Integer frame: s){
            improc = istack.getProcessor(frame).convertToFloat();
            cnets = s.getCoordinates(frame);
            double l = s.TYPE==Snake.CLOSED_SNAKE?
                    s.findLength(frame) + calculateDistance(cnets.get(0), cnets.get(cnets.size()-1)):
                    s.findLength(frame);

         	double [] cm = getCenter(cnets);
         	System.out.println("running fits: " +  cm[0]+ " " + cm[1] + " " + cm[2]);
         	centerX += cm[0];
         	centerY += cm[1];
         	radius += cm[2];
         	fs++;
        }
        centerX /= fs;
        centerY /= fs;
        radius /= fs;
        GenericDialog gd = new GenericDialog("Constants");
        // read in the width of the kymograph ribbon and cell center
        LINE_WIDTH=4;
        dx = 0;
        dt = 0;
        nstks=0;
        tfit=1;
    

        gd.addNumericField("xCM, X-coordinate of cell center (px): ", centerX, 2);
        gd.addNumericField("yCM, Y-coordinate of cell center (px): ", centerY, 2);
        gd.addNumericField("dx, image resolution (nm/px): ", dx,2);
        gd.addNumericField("N, number of frames: ", nstks, 2);
        gd.addNumericField("Rotate data by 180 degrees? (0=no, 1=yes):", dtheta, 1);    

        gd.showDialog();

        if (gd.wasCanceled())   return;
        
        centerX = (int)gd.getNextNumber();
        centerY = (int)gd.getNextNumber();
        dx = (double)gd.getNextNumber();
        nstks = (int)gd.getNextNumber();
        dtheta = (int)gd.getNextNumber(); 

        
        LINE_WIDTH = back-front;
        rads = new double[nstks][360];
        for(int i=0;i<nstks;i++)
        	for(int j=0; j< 360;j++)
        			rads[i][j]=0;
                       
        int MAX_WIDTH = 0;
        for(Integer frame: s){
            improc = istack.getProcessor(frame).convertToFloat();
            cnets = s.getCoordinates(frame);
            if(frame>maxF) maxF = frame;

            double l = s.TYPE==Snake.CLOSED_SNAKE?
                s.findLength(frame) + calculateDistance(cnets.get(0), cnets.get(cnets.size()-1)):
                s.findLength(frame);

            kymo_pixels.clear();
            double[] yvalues = createKimograph(cnets, improc, l, s.TYPE,centerX,centerY, frame, back, dtheta);
            for(int j=0; j<360; j++) rads[frame-1][j]=yvalues[j];        
        }
        writeOut(nstks);
        
    }

    public double[] getCenter(ArrayList<double[]> pts){
 	   int N = pts.size(); // number of pts in our snake
 	   int reps = 100;
 	   int j;
 	   Random generator = new Random();
 	   double xo,yo,r;
 	   double Axo,Ayo,Ar;
 	   double [] p1, p2,p3;
 	   double ma,mb;
 	   Axo=0;
 	   Ayo=0;
 	   Ar=0;	   
 	   
 	   for(int i=0; i<reps;i++){
 		   // do 100 tries to fit
 		   
 		   j = (int)(N/3*generator.nextDouble());
 		   p1 =  pts.get(j);
 		   j = (int)(N/3*generator.nextDouble() + N/3);
 		   p2 =  pts.get(j);
 		   j = (int)(N/3*generator.nextDouble() + 2*N/3);
 		   p3 =  pts.get(j);
 		   
 		   ma = (p2[1]-p1[1])/(p2[0]-p1[0]);
 		   mb = (p2[1]-p3[1])/(p2[0]-p3[0]);
 		   
 		   xo = ma*mb*(p1[1]-p3[1]) + mb*(p1[0]+p2[0])-ma*(p2[0]+p3[0]);
 		   xo /= 2*(mb-ma);
 		   
 		   yo = -1/ma *(xo - (p1[0]+p2[0])/2) + (p1[1]+p2[1])/2;
 		   
 		   r = (p1[0]-xo)*(p1[0]-xo) + (p1[1]-yo)*(p1[1]-yo);
 		   r = Math.sqrt(r);
 		   
 		   Axo += xo/reps;
 		   Ayo += yo/reps;
 		   Ar += r/reps;		   
 	   }
 	   double[] cms = new double[3];
 	   cms[0] = Axo;
 	   cms[1] = Ayo;
 	   cms[2] = Ar;
 	   return(cms);
    }


    private void writeOut(int maxF) {

        
            try{
                if(bw==null){
                    FileDialog fd = new FileDialog(ij.IJ.getInstance(),"Save CSV File",FileDialog.SAVE);
                    fd.setFile("RadiusMatrix.txt");
                    fd.setVisible(true);
                    String fname = fd.getFile();
                    String dirname = fd.getDirectory();
                    String outname = dirname +  fname;
                    bw = new BufferedWriter(new FileWriter(outname));
                }
                                      
             	               
                 for(int j=0;j<360;j++){
                 	for(int i=0; i< maxF;i++){
                	 	 bw.write(rads[i][j]*dx/1000.0 + "\t");
                	 }
                	 bw.write("\n");
                 }
                 
            }catch(Exception e){
                 System.out.println("problems");
                e.printStackTrace();
            }
            
            try{
                bw.close();
            }catch(Exception e){
                //pass
            }
    }
   
    public double[] createKimograph(ArrayList<double[]> pts,ImageProcessor ip, double l, int type,double centerX, double centerY, int frame, double back, int dtheta){
        int N = pts.size();

        // number of points to consider (N  in snake, so N-1 intervals if the snake is open, N if closed)
        int n = type==Snake.OPEN_SNAKE?N-1:N;

        double[] xvalues = new double[360];
        double[] yvalues = new double[360];
        double[] angle = new double[n];      
        double[] tangle = new double[n]; 
        double[][] coors = new double[n][2];
        int j;
        double iCHANGE = LINE_WIDTH/((int)LINE_WIDTH + 1); // spacing inwards (radially) from the snake
        int Ni = (int)((LINE_WIDTH+iCHANGE)/iCHANGE);
        double[] slice;
        slice = new double[Ni];
        
        for(int p=0;p<Ni;p++) slice[p]=0;     
        for(int k=0; k<360 ;k++)
           	{
        		yvalues[k]=0;
        		xvalues[k]=k;
        		kymo_pixels.add(slice);
           	}

        // get all the angles calculated
        for(int i=0;i<n;i++)
        {
        	double[] now = pts.get(i);
        	tangle[i] = getTheta(now,centerX,centerY, dtheta);
        }
        
        boolean flip=false;
        for(int i=0;i<n-1;i++)
        	{
        		double dift = tangle[i+1]-tangle[i];
        		if(Math.abs(dift)>180) flip = true;
        	}
        if(flip){
        	 for(int i=0;i<n;i++)
        	 {
        		 if(tangle[i]>180) tangle[i] = tangle[i]-360;
        	 }
        }
        
        //sort the files here
        int minT;
        for(int i=0;i<n;i++){
        	minT = 0;
        	for(int ji=0;ji<n;ji++){
        		if(tangle[ji]<tangle[minT]) minT=ji;
        	}
        	angle[i] = tangle[minT];
        	coors[i]= pts.get(minT);     
        	tangle[minT]=500;
        }
        //now have ordered coordinates with increasing theta
                
        double[] top;
        double[] bottom;
        for(double theta = Math.ceil(angle[0]);theta <angle[n-1];theta+=1.0)
        {
        	int i = 0;
        	while(theta>angle[i])i++;
        	bottom = coors[i-1];
        	top = coors[i];
        	int k = (int)theta;
        	if(k<0) k+=360;
        	// select two points on the snake bordering the angle we want
    
        	yvalues[k] = collectBox(bottom, top, ip, centerX,centerY, theta, frame,back,dtheta); //Average intensity of this box
        
        }             
       
        return yvalues;
    }

    
    public double getTheta(double[] pt, double centerX, double centerY, int dtheta){
    	 double py = (pt[1] - centerY);
    	 double px = (pt[0] - centerX);
    	 double theta;  // should return angle between 0 and pi/2 (positive)
    	 if(px==0){
    		 theta = 3.141592654/2.0;
    	 }else if(py==0){
    		 theta = 0;
    	 }else{
    		 theta = Math.atan(py/px);
    	 }
    	 theta *= 180 / 3.141592654; // put into degrees
    	 
         if(py>0){// in the top half
         	if(theta<0) theta +=180;
         }else{ // lower half
         	if(px<0) {    //neg x and y, positive theta		
         		theta = 180+theta;
         	}else{ //neg y only, neg theta
         		theta =360+theta;
         	}
         }	     
         theta += dtheta*180;
         if(theta >=360) theta -=360;
         if(theta<0) theta +=360;

    	return(theta);
    }
    
    
    
    public double calculateDistance(double[] a, double[] b){
    	// calculates the scalar distance between two input points, a and b
        double r = 0;
        for(int i = 0; i<a.length; i++)
            r += Math.pow(a[i] - b[i], 2);

        return Math.sqrt(r);

    }
    
    
    
    public double collectBox(double[] p1, double[] p2, ImageProcessor ip, double centerX, double centerY, double theta, int frame, double back, int dtheta){ // measures the average intensity of a snake segment
// bottom point, top point....
        //principle axis
        double lx = p2[0] - p1[0]; // cos theta
        double ly = p2[1] - p1[1]; //sin theta
        double lx_sq = Math.pow(lx,2);
        double ly_sq = Math.pow(ly,2);
        double length = Math.sqrt(lx_sq+ly_sq); //distance between the two end points
        double costh = lx/length;
        double sinth = ly/length;
        double thb = getTheta(p1, centerX, centerY, dtheta);
        double tht = getTheta(p2, centerX, centerY, dtheta);
        
        double diffth = tht-thb; // difference between two angles
        double diffs = theta-thb;
        if(diffth>360) diffth -= 360;
        if(diffth<0) diffth += 360;
        if(diffs>360) diffs -= 360;
        if(diffs<0) diffs += 360;
        
        double ratio = diffs/diffth; // fraction of length to 'walk' between two points
        
        double mx = p1[0] + ratio * length * costh;
        double my = p1[1] + ratio * length * sinth;        
        /*
        // take avg b/w our two points
        double mx = (p1[0]+p2[0])/2.0;
        double my = (p1[1]+p2[1])/2.0;
        */
        double rx = mx - centerX;
        double ry = my - centerY;
        double rlength = Math.sqrt(rx*rx+ry*ry);

        return rlength; // average value returned
        //return value; // total value returned


    }
}